
const readxml = require("./xml");
const readCht = require("./cht");
const readline = require("readline");
const { argv, stdin, stdout } = require("process");
const { readFile, writeFile, access } = require("fs/promises");

const codeA = "A".charCodeAt();
const codeZ = "Z".charCodeAt();
const codea = "a".charCodeAt();
const codez = "z".charCodeAt();
const code0 = "0".charCodeAt();
const code9 = "9".charCodeAt();

const candylen = 10;

const question = (con, text) => new Promise( (ok, err) => {
    con.question( text, ans => ok(ans.length == 0 ? "1" : ans) );
});

const isSymbol = c => {
    const code = c.charCodeAt();
    return code < code0 || (code9 < code && code < codeA) || (codeZ < code && code < codea) || (codez < code && code < 128);
}

const split = tr => {
    const id = tr.slice(0, 4);
    const tags = tr.match(/\([^\)]*\)/g);
    const names = tr.slice(7).replace(/\([^\)]*\)/g, "").replace(/&[^;]*;/g, "").trim();

    let [words, chineser] = names.split(" ").reduce( (r,v) => {
        let chars = v.split("");
        let [word, isChinese] = chars.reduce( (r, v) => {
            if( v.charCodeAt() > 128 )
            {
                r[1] = true;
                r[0].push( v )
            }
            else if( r[1] == true || false == isSymbol(v) )
                r[0].push( v );
            return r;
        }, [ [], false]);
        if( isChinese )
        {
            r[1].push( word.join("") );
        }
        else if( /^[0-9]+$/.test(v) )
        {
            if( r[0].length == 0 ) r[1].push( word.join("") );
            else r[0].push( word.join("") );
        }
        else if( word.length > 0 ) 
        {
            r[0].push( word.join("") );
        }
        return r;
    }, [ [], [] ])
    
    const chinese = chineser.join(" ");

    return { id, tags, name: chinese, words, text: tr.trim() };
}

const scoring = (standard, target) => {
    let score = 0;

    // 1 分词(在调用者处完成)
    let [skeys, tkeys] = [standard, target];

    // 2 匹配，找到+100，顺序一致+100
    let order = -1;
    let smatched = 0, tmatched = 0;
    for( let skey of skeys )
    {
        let res = tkeys.indexOf(skey);
        if( res >= 0 ) // 匹配了
        {
            score += 100;
            if( res > order )//顺序一致
                score += 100;
            ++smatched;
            ++tmatched;
            order = res;
        }
    }

    // 4 完成所有匹配，计算剩余词汇数量
    // 5 standard无剩余词汇+1000
    if( smatched == skeys.length )
    {
        score += 1000;
    }
    else
    {
        let miss = skeys.length - smatched;
        if( miss <= 2 ) score += 60 / miss;
    }

    // 6 target无剩余词汇，满足5则+1000，不满足+500
    if( tmatched == tkeys.length )
    {
        score += smatched == skeys.length ? 1000 : 500;
    }

    // 7 target有剩余词汇，剩余词汇数量>4不计分，否则+120/数量
    else 
    {
        let miss = tkeys.length - tmatched;
        if( miss <= 4 ) score += 60 / miss;
    }

    return score;
}

const loadInfo = async () => {
    const datfile = await readFile("./list.dat");
    const records = readxml( datfile.toString("utf-8") );
    const inform = rec => Object.assign( split(rec.attributes.name), {serial: rec.children[1].attributes.serial||"????"} );
    return records.children.filter( rec => rec.name=="game" && rec.attributes.name.charAt()!="x" ).map( inform );
}

const loadIDs = async () => {
    const idfile = await readFile("./gbaid.txt");
    const rom = name => {
        const rname = name.trim();
        return name.length > 0 ? split(name) : null;
    }
    return idfile.toString("utf-8").split("\n").map( rom ).filter( s=>s!=null );
}

const findnear = (standard, list) => {
    let res = [];
    let lowerbound = 0;
    for( let info of list ){
        let score = scoring( standard.words, info.words );
        if( score <= 150 ) continue;
        if( res.length < candylen )
        {
            res.push( {info, score} );
            if( lowerbound == 0 ) lowerbound = score;
            else if( score < lowerbound ) lowerbound = score;
        }
        else if( score > lowerbound )
        {
            let min = 10000000;
            let done = false;
            for( let store of res ){
                if( store.score == lowerbound && !done ){
                    store.info = info;
                    store.score = score;
                    done = true;
                }
                if( store.score < min ) min = store.score;
            }
            lowerbound = min;
        }
    }
    return res.sort( (a,b) => a.score != b.score ? b.score - a.score : a.info.id.localeCompare(b.info.id) );
}

const start = async () => 
{
    const base = await loadIDs();
    const sys = await loadInfo();

    let id = argv.length > 2 ? parseInt( argv[2] ) : 1;
    let index = base.findIndex( x => parseInt(x.id) == id );
    if( index < 0 ) index = 0;

    const con = readline.createInterface( {input: stdin, output: stdout} );
    let edit = {};
    while( true )
    {
        const standard = base[index];
        console.log( `正在匹配：${standard.text}` );
        const results = findnear( standard, sys );
        results.forEach( (r, i) => {
            console.log( `${i+1} - ${r.info.text}(${r.info.serial})[${r.score}]` );
        } );
        const answer = await question( con, `输入你的选择[1]：` );
        console.log( `你的答案：${answer}` );
 
        if( answer == "exit" ) break;
        else if( /^[0-9]+$/g.test(answer) )
        {
            let pos = parseInt(answer) - 1;
            if( answer == "0" ) {
                let auto = await question( con, `您选择自行输入ID：` );
                let autoid = parseInt(auto);
                if( !isNaN(autoid) )
                    pos = sys.findIndex( x => parseInt(x.id) == autoid );

                if( pos >= 0 ) 
                {
                    auto = await question( con, `使用${sys[pos].text}?` );
                    if( auto != "yes" ) pos = -1;
                    else results[pos] = {info: sys[pos], score: 3000};
                }
            }
            if( !results[pos] ) console.log( `输入项无效！` );
            else
            {
                edit[ standard.id ] = results[pos].info.serial;
                if( index + 1 >= base.length ) break;
                else ++index;
            }
        }
        else console.log( `请用心输入` );
    }
    console.log( edit );
    con.close();
}

start();
/*(async () => {
  const a=await loadIDs();
  const b=await loadInfo();
  console.log( a[80], b[79] );
  console.log( findnear(a[80], b) );
})()*/
